# SPDX-FileCopyrightText: Magenta ApS
# SPDX-License-Identifier: MPL-2.0
import json  # pragma: no cover
from typing import Tuple  # pragma: no cover
from uuid import UUID  # pragma: no cover

import click  # pragma: no cover

from . import AsyncDARClient  # pragma: no cover
from fastramqpi.ra_utils.async_to_sync import async_to_sync  # pragma: no cover


@click.command()
@click.option(
    "uuids",
    "--uuid",
    type=click.UUID,
    multiple=True,
    required=True,
    help="DAR UUIDs to lookup",
)
@async_to_sync
async def cli(uuids: Tuple[UUID]) -> None:  # pragma: no cover
    darclient = AsyncDARClient()
    async with darclient:
        if not await darclient.healthcheck():
            raise click.ClickException("Unable to establish connection to DAR")

        results, missing = await darclient.fetch(set(uuids))
        print("Found:", json.dumps(list(results.values()), indent=4))
        print("Missing:", json.dumps(list(map(str, missing)), indent=4))


if __name__ == "__main__":  # pragma: no cover
    cli()



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Any
from typing import Optional


class AttrDict(dict):
    """Enable dot.notation access for a dict object.

    Example:
        ```Python
        script_result = AttrDict({"exit_code": 0})
        assert script_result.exit_code == 0
        ```
    """

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__  # type: ignore
    __delattr__ = dict.__delitem__  # type: ignore


def attrdict(*args: Any, **kwargs: Optional[Any]) -> AttrDict:
    """Constructor method for `AttrDict`s.

    Example:
        ```Python
        script_result = attrdict({"exit_code": 0})
        assert script_result.exit_code == 0
        ```
    """

    return AttrDict(*args, **kwargs)



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import sqlalchemy
from pydantic import BaseSettings
from sqlalchemy import MetaData
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio import create_async_engine

from fastramqpi.config import DatabaseSettings


def create_engine(
    user: str,
    password: str,
    host: str,
    port: int,
    name: str,
) -> AsyncEngine:
    url = f"postgresql+psycopg://{user}:{password}@{host}:{port}/{name}"
    return create_async_engine(url)


def create_sessionmaker(engine: AsyncEngine) -> async_sessionmaker:
    return async_sessionmaker(engine)


def run_upgrade(database_metadata: MetaData) -> None:
    """Apply alembic migrations (in the future).

    TODO: when we implement alembic, the caller should no longer pass us their
    metadata; all of the table definitions will be apparent from the migrations in each
    integration (which will be available to alembic).
    In fact, we probably don't even want to run migrations during app startup, but
    delegate it to a CLI interface or similar, which can be run externally to the main
    process. In this case, we still need to implement programmatic migrations, to
    implement the CLI in python and run migrations from the test suite.

    These cookbooks could be useful (and are used in OS2mo's setup):
    https://alembic.sqlalchemy.org/en/latest/cookbook.html#programmatic-api-use-connection-sharing-with-asyncio
    https://alembic.sqlalchemy.org/en/latest/cookbook.html#connection-sharing
    https://alembic.sqlalchemy.org/en/latest/cookbook.html#asyncio-recipe
    """

    class Settings(DatabaseSettings, BaseSettings):
        """Load database variables without depending on the integration's settings."""

        class Config:
            env_prefix = "FASTRAMQPI__DATABASE__"

    # TODO: when we implement alembic the whole interface will be async and we can use
    # our own create_engine() from above.
    db = Settings()
    url = f"postgresql+psycopg://{db.user}:{db.password}@{db.host}:{db.port}/{db.name}"
    engine = sqlalchemy.create_engine(url)
    # Create all tables in the metadata, ignoring tables already present in the
    # database. A proper migration tool, such as alembic, is more appropriate.
    # https://docs.sqlalchemy.org/en/20/tutorial/metadata.html#emitting-ddl-to-the-database
    with engine.begin() as connection:
        database_metadata.create_all(connection)



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""FastAPI + RAMQP Framework."""
from contextlib import asynccontextmanager
from functools import partial
from typing import Any
from typing import AsyncContextManager
from typing import AsyncGenerator
from typing import cast
from typing import Protocol
from typing import Type

import httpx
from authlib.integrations.httpx_client import AsyncOAuth2Client
from sqlalchemy import MetaData

from . import database
from .app import FastAPIIntegrationSystem
from .config import ClientSettings
from .config import Settings
from .context import Context
from .raclients.graph.client import GraphQLClient as LegacyGraphQLClient
from .raclients.modelclient.mo import ModelClient as LegacyModelClient
from .ramqp.mo import MOAMQPSystem


def construct_legacy_clients(
    graphql_version: int,
    settings: ClientSettings,
) -> tuple[LegacyGraphQLClient, LegacyModelClient]:
    """Construct legacy clients froms settings.

    Args:
        settings: Integration settings module.

    Returns:
        Tuple with PersistentGraphQLClient and ModelClient.
    """
    # DEPRECATED: ariadne-codegen is the preferred way to interface with GraphQL
    gql_client = LegacyGraphQLClient(
        url=f"{settings.mo_url}/graphql/v{graphql_version}",
        client_id=settings.client_id,
        client_secret=settings.client_secret.get_secret_value(),
        auth_realm=settings.auth_realm,
        auth_server=settings.auth_server,
        execute_timeout=settings.graphql_timeout,
        httpx_client_kwargs={"timeout": settings.graphql_timeout},
    )
    # DEPRECATED: GraphQL is the preferred way to interface with OS2mo
    model_client = LegacyModelClient(
        client_id=settings.client_id,
        client_secret=settings.client_secret.get_secret_value(),
        auth_realm=settings.auth_realm,
        auth_server=settings.auth_server,
        base_url=settings.mo_url,
    )
    return gql_client, model_client


class GraphQLClientProtocol(AsyncContextManager, Protocol):
    def __init__(
        self,
        url: str = "",
        headers: dict[str, str] | None = None,
        http_client: httpx.AsyncClient | None = None,
        ws_url: Any = None,
        ws_headers: Any = None,
        ws_origin: Any = None,
        ws_connection_init_payload: Any = None,
    ) -> None:
        ...  # pragma: no cover


class FastRAMQPI(FastAPIIntegrationSystem):
    """FastRAMQPI (FastAPI + RAMQP) combined-system.

    Motivated by a lot a shared code between our AMQP integrations.
    """

    def __init__(
        self,
        application_name: str,
        settings: Settings,
        graphql_version: int,
        graphql_client_cls: Type[GraphQLClientProtocol] | None = None,
        database_metadata: MetaData | None = None,
    ) -> None:
        super().__init__(application_name, settings)

        # Setup AMQPSystem
        amqp_settings = cast(Settings, self.settings).amqp
        amqp_settings.queue_prefix = amqp_settings.queue_prefix or application_name
        self.amqpsystem = MOAMQPSystem(
            settings=amqp_settings, context=self.get_context()
        )
        # Let AMQPSystems lifespan follow ASGI lifespan. The AMQPSystem must be started
        # last (priority 1000) so all the dependency-injectable objects are set up and
        # available before the AMQP message handlers start receiving messages. Since
        # lifespan managers are shutdown in reverse order, this also ensures that all
        # messages handlers have finished before shutting down their dependencies.
        self.add_lifespan_manager(self.amqpsystem, priority=1000)

        async def healthcheck_amqp(context: Context) -> bool:
            """AMQP Healthcheck wrapper.

            Args:
                context: unused context dict.

            Returns:
                Whether the AMQPSystem is OK.
            """
            amqpsystem = context["amqpsystem"]
            return cast(bool, amqpsystem.healthcheck())

        self.add_healthcheck(name="AMQP", healthcheck=healthcheck_amqp)
        self._context["amqpsystem"] = self.amqpsystem

        # Setup database
        if database_metadata is not None:
            assert settings.database is not None, "database settings missing"
            database_engine = database.create_engine(
                user=settings.database.user,
                password=settings.database.password,
                host=settings.database.host,
                port=settings.database.port,
                name=settings.database.name,
            )
            database.run_upgrade(database_metadata)
            self._context["sessionmaker"] = database.create_sessionmaker(
                database_engine
            )

        # Authenticated HTTPX Client
        mo_client = AsyncOAuth2Client(
            base_url=settings.mo_url,
            client_id=settings.client_id,
            client_secret=settings.client_secret.get_secret_value(),
            grant_type="client_credentials",
            # TODO: We should take a full token URL instead of hard-coding Keycloak's
            # URL scheme. Let's wait until the legacy clients are removed.
            token_endpoint=f"{settings.auth_server}/realms/{settings.auth_realm}/protocol/openid-connect/token",
            # TODO (https://github.com/lepture/authlib/issues/531): Hack to enable
            # automatic fetching of token on first call, instead of only refreshing.
            token={"expires_at": -1, "access_token": ""},
            timeout=settings.graphql_timeout,
        )

        @asynccontextmanager
        async def mo_client_manager(context: Context) -> AsyncGenerator[None, None]:
            async with mo_client as client:
                context["mo_client"] = client
                yield

        self.add_lifespan_manager(
            cast(AsyncContextManager, partial(mo_client_manager, self._context)()),
            priority=100,
        )

        # GraphQL Client
        if graphql_client_cls is not None:

            @asynccontextmanager
            async def graphql_client_manager(
                context: Context,
            ) -> AsyncGenerator[None, None]:
                graphql_client = graphql_client_cls(
                    url=f"{settings.mo_url}/graphql/v{graphql_version}",
                    http_client=mo_client,
                )
                async with graphql_client as client:
                    context["graphql_client"] = client
                    yield

            self.add_lifespan_manager(
                cast(
                    AsyncContextManager,
                    partial(graphql_client_manager, self._context)(),
                ),
                priority=200,
            )

        # Prepare legacy clients
        legacy_graphql_client, legacy_model_client = construct_legacy_clients(
            graphql_version=graphql_version,
            settings=cast(ClientSettings, self.settings),
        )
        # Expose legacy GraphQL connection (gql_client)
        self._context["legacy_graphql_client"] = legacy_graphql_client

        @asynccontextmanager
        async def legacy_graphql_session(
            context: Context,
        ) -> AsyncGenerator[None, None]:
            async with context["legacy_graphql_client"] as session:
                context["legacy_graphql_session"] = session
                yield

        self.add_lifespan_manager(
            cast(AsyncContextManager, partial(legacy_graphql_session, self._context)()),
            priority=300,
        )
        # Expose legacy Service API connection (model_client)
        self.add_lifespan_manager(legacy_model_client, priority=400)
        self._context["legacy_model_client"] = legacy_model_client

    def get_amqpsystem(self) -> MOAMQPSystem:
        """Return the contained MOAMQPSystem.

        Returns:
            MOAQMPSystem.
        """
        return self.amqpsystem



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2022 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
from typing import Any
from typing import Dict
from typing import Protocol
from uuid import UUID

from fastapi.encoders import jsonable_encoder
from httpx import AsyncClient

from .base import ModelClientBase


class LoraBase(Protocol):
    @property
    def uuid(self) -> UUID:  # pragma: no cover
        ...


class ModelClient(ModelClientBase[LoraBase]):
    """LoRa ModelClient.

    Example usage::

        async with ModelClient(
            base_url="http://mox:8080",
            client_id="AzureDiamond",
            client_secret="hunter2",
            auth_server=parse_obj_as(AnyHttpUrl,"https://keycloak.example.org/auth"),
            auth_realm="mordor",
        ) as client:
            r = await client.upload(objects)
    """

    upload_http_method = "PUT"
    path_map: Dict[str, str] = {
        "Facet": "/klassifikation/facet",
        "ITSystem": "/organisation/itsystem",
        "Klasse": "/klassifikation/klasse",
        "Organisation": "/organisation/organisation",
    }
    async_httpx_client_class = AsyncClient

    def get_object_url(self, obj: LoraBase, *args: Any, **kwargs: Any) -> str:
        return "{}/{}".format(self.path_map[type(obj).__name__], obj.uuid)

    def get_object_json(self, obj: LoraBase, *args: Any, **kwargs: Any) -> Any:
        # TODO: Pending https://github.com/samuelcolvin/pydantic/pull/2231. For now,
        #  UUID is included in the model, and has to be excluded when converted to JSON.
        return jsonable_encoder(
            obj=obj, by_alias=True, exclude={"uuid"}, exclude_none=True
        )



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
from functools import cache
from typing import Annotated
from typing import Any
from typing import AsyncIterator
from typing import Callable

from authlib.integrations.httpx_client import AsyncOAuth2Client
from fastapi import Depends
from gql.client import AsyncClientSession
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.ext.asyncio import AsyncSession

from .raclients.graph.client import GraphQLClient as _GraphQLClient
from .raclients.modelclient.mo import ModelClient as _ModelClient
from .ramqp.depends import from_context
from .ramqp.mo import MOAMQPSystem as _MOAMQPSystem


@cache
def from_user_context(field: str) -> Callable[..., Any]:
    """Construct a Callable which extracts 'field' from the FastRAMQPI user context.

    Args:
        field: The field to extract.

    Returns:
        A callable which extracts 'field' from the FastRAMQPI user context.
    """

    def inner(user_context: UserContext) -> Any:
        return user_context[field]

    return inner


MOAMQPSystem = Annotated[_MOAMQPSystem, Depends(from_context("amqpsystem"))]


async def get_session(sessionmaker: "Sessionmaker") -> AsyncIterator[AsyncSession]:
    async with sessionmaker() as session, session.begin():
        yield session


Sessionmaker = Annotated[async_sessionmaker, Depends(from_context("sessionmaker"))]
Session = Annotated[AsyncSession, Depends(get_session)]

LegacyGraphQLClient = Annotated[
    _GraphQLClient, Depends(from_context("legacy_graphql_client"))
]
LegacyGraphQLSession = Annotated[
    AsyncClientSession, Depends(from_context("legacy_graphql_session"))
]
LegacyModelClient = Annotated[
    _ModelClient, Depends(from_context("legacy_model_client"))
]

MOClient = Annotated[AsyncOAuth2Client, Depends(from_context("mo_client"))]

UserContext = Annotated[dict[str, Any], Depends(from_context("user_context"))]



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""FastAPI Framework."""
from contextlib import asynccontextmanager
from contextlib import AsyncExitStack
from contextlib import suppress
from functools import partial
from typing import Any
from typing import AsyncContextManager
from typing import AsyncIterator

import structlog
from fastapi import APIRouter
from fastapi import FastAPI
from fastapi import Request
from fastapi.responses import JSONResponse
from prometheus_client import Info
from prometheus_fastapi_instrumentator import Instrumentator
from starlette.status import HTTP_200_OK
from starlette.status import HTTP_503_SERVICE_UNAVAILABLE

from .config import FastAPIIntegrationSystemSettings
from .context import Context
from .context import HealthcheckFunction
from fastramqpi.logging import configure_logging

logger = structlog.get_logger()
fastapi_router = APIRouter()
build_information = Info("build_information", "Build information")


def update_build_information(version: str, build_hash: str) -> None:
    """Update build information.

    Args:
        version: The version to set.
        build_hash: The build hash to set.

    Returns:
        None.
    """
    build_information.info(
        {
            "version": version,
            "hash": build_hash,
        }
    )


@fastapi_router.get("/")
async def index(request: Request) -> dict[str, str]:
    """Endpoint to return name of integration."""
    context: dict[str, Any] = request.state.context
    return {"name": context["name"]}


@fastapi_router.get(
    "/health/live",
    status_code=HTTP_200_OK,
    responses={
        "200": {"description": "Ready"},
        "503": {"description": "Not ready"},
    },
)
@fastapi_router.get(
    "/health/ready",
    status_code=HTTP_200_OK,
    responses={
        "200": {"description": "Ready"},
        "503": {"description": "Not ready"},
    },
)
async def healthcheck_probe(request: Request) -> JSONResponse:
    """Kubernetes healthcheck probe function.

    Kubernetes defines the probes as follows:
      - The livenessProbe is used to decide whether the integration should be
        restarted or not.
      - The readinessProbe is used to decide whether the integration should
        receive traffic or not.
    and the implementation of the probes:
      - The livenessProbe can be implemented however you would like.
      - The readinessProbe should whether check back-end services are
        available.

    Using the same probe implementation for both allows us to configure both a
    readinessProbe and a livenessProbe for the integration, for instance such
    that we can stop sending traffic to an integration while it has a temporary
    outage, but that we will only resort to actually restarting the integration
    if the situation does not seem as temporary as first assumed.
    """
    status_code = HTTP_200_OK

    context: Context = request.state.context
    healthchecks: dict[str, HealthcheckFunction] = request.app.state.healthchecks

    async def check(healthcheck: HealthcheckFunction) -> bool:
        with suppress(Exception):
            return await healthcheck(context)
        return False

    healthstatus = {
        name: await check(healthcheck) for name, healthcheck in healthchecks.items()
    }
    if not all(healthstatus.values()):
        status_code = HTTP_503_SERVICE_UNAVAILABLE

    return JSONResponse(content=healthstatus, status_code=status_code)


@asynccontextmanager
async def _lifespan(app: FastAPI, context: Context) -> AsyncIterator[dict]:
    """ASGI lifespan context handler.

    Runs all the configured lifespan managers according to their priority.

    Returns:
        None
    """
    async with AsyncExitStack() as stack:
        lifespan_managers = context["lifespan_managers"]
        for _, priority_set in sorted(lifespan_managers.items()):
            for lifespan_manager in priority_set:
                await stack.enter_async_context(lifespan_manager)
        yield {
            "context": context,
        }


def enable_debugging() -> None:  # pragma: no cover
    import debugpy  # type: ignore

    logger.debug("Enabling debugging", port=5678)
    debugpy.listen(("0.0.0.0", 5678))


class FastAPIIntegrationSystem:
    """FastAPI-based integration framework.

    Motivated by a lot of shared code between our integrations.
    """

    def __init__(
        self, application_name: str, settings: FastAPIIntegrationSystemSettings
    ) -> None:
        super().__init__()
        self.settings = settings

        configure_logging(self.settings.log_level, self.settings.json_logs)

        if self.settings.dap:  # pragma: no cover
            enable_debugging()

        # Setup shared context
        self._context: Context = {
            "name": application_name,
            "settings": self.settings,
            "lifespan_managers": {},
            "user_context": {},
        }

        # Setup FastAPI
        app = FastAPI(
            title=application_name,
            version=self.settings.commit_tag,
            contact={
                "name": "Magenta Aps",
                "url": "https://www.magenta.dk/",
                "email": "info@magenta.dk",
            },
            license_info={
                "name": "MPL-2.0",
                "url": "https://www.mozilla.org/en-US/MPL/2.0/",
            },
            lifespan=partial(_lifespan, context=self._context),
        )
        app.state.context = self._context
        app.state.healthchecks = {}
        app.include_router(fastapi_router)
        # Expose Metrics
        if self.settings.enable_metrics:
            # Update metrics info
            update_build_information(
                version=self.settings.commit_tag, build_hash=self.settings.commit_sha
            )

            instrumentator = Instrumentator()
            self._context["instrumentator"] = instrumentator
            instrumentator.instrument(app).expose(app)
        self.app = app
        self._context["app"] = self.app

    def add_lifespan_manager(
        self, manager: AsyncContextManager, priority: int = 1000
    ) -> None:
        """Add the provided life-cycle manager to the ASGI lifespan context.

        Args:
            manager: The manager to add.
            priority: The priority of the manager, lowest priorities are run first.

        Returns:
            None
        """

        priority_set = self._context["lifespan_managers"].setdefault(priority, set())
        priority_set.add(manager)

    def add_healthcheck(self, name: str, healthcheck: HealthcheckFunction) -> None:
        """Add the provided healthcheck to the Kubernetes readiness probe.

        Args:
            name: Name of the healthcheck to add.
            healthcheck: The healthcheck callback function.

        Raises:
            ValueError: If the name has already been used.

        Returns:
            None
        """
        if name in self.app.state.healthchecks:
            raise ValueError("Name already used")
        self.app.state.healthchecks[name] = healthcheck

    def add_context(self, **kwargs: Any) -> None:
        """Add the provided key-value pair to the user-context.

        The added key-value pair will be available under context["user_context"].

        Args:
            key: The key to add under.
            value: The value to add.

        Returns:
            None
        """
        self._context["user_context"].update(**kwargs)

    def get_context(self) -> Context:
        """Return the contained context.

        Returns:
            The contained context.
        """
        return self._context

    def get_app(self) -> FastAPI:
        """Return the contained FastAPI application.

        Returns:
            FastAPI application.
        """
        return self.app



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
import ast
from datetime import datetime
from importlib.metadata import version
from typing import Any
from typing import cast
from zoneinfo import ZoneInfo

from ariadne_codegen.plugins.base import Plugin
from graphql import GraphQLInputField
from graphql import Undefined
from pydantic.datetime_parse import parse_datetime
from pydantic.datetime_parse import StrBytesIntFloat


MO_TZ = ZoneInfo("Europe/Copenhagen")


def parse_graphql_datetime(value: StrBytesIntFloat | datetime) -> datetime:
    """Parse OS2mo GraphQL datetime to Python object.

    Even though ISO 8601 promises "unique and unambiguous" representations of
    datetimes, it is impossible to unambiguously represent timestamp in a
    specific time zone using only a UTC offset.

    For example, the 31st October 2010 at 00:00 in Copenhagen and Cairo is both
    represented as the timestamp `2010-10-31T00:00:00+02:00`. However, one day
    later in Copenhagen is represented as `2010-11-01T00:00:00+01:00` -- now
    with a UTC offset of 01:00 because of daylight saving time -- whereas the
    timestamp in Cairo is `2010-11-01T00:00:00+02:00`.

    ISO 8601 is thus inherently a broken standard. Until RFC 9557 introduces
    proper timestamps with time zone locations, the best we can do is assume
    that everyone observes Copenhagen time.
    """
    # Datetime objects can have proper time zone information, so we refrain
    # from modifying it. Values received from OS2mo will always be strings.
    if isinstance(value, datetime):
        return value
    dt = parse_datetime(value)
    # Only assume Copenhagen time if UTC offset matches Copenhagen
    if dt.utcoffset() != dt.astimezone(MO_TZ).utcoffset():
        return dt
    return dt.replace(tzinfo=MO_TZ)


def _is_ast_annotation_optional(annotation: ast.expr) -> bool:
    match annotation:
        # This case handles the `B[A]` syntax
        case ast.Subscript():
            # To be optional, `B` in the syntax must be the string "Optional"
            value = annotation.value
            # Not a name --> not a match
            if not isinstance(value, ast.Name):
                return False
            return value.id == "Optional"

        # This case handles the `A op B` syntax
        case ast.BinOp():
            # To be optional, `B` in the syntax must be None and op must be "|"
            op = annotation.op
            # if op is not "|" this is not a match
            if not isinstance(op, ast.BitOr):
                return False
            match annotation.right:
                case ast.Name():
                    return annotation.right.id == "None"
                case ast.Constant():
                    return annotation.right.value is None
                case _:
                    return False

        # This case handles the stringified `"Optional[A]"` syntax
        # This case handles the stringified `"A | None"` syntax
        case ast.Name():
            return "Optional[" in annotation.id or "| None" in annotation.id

        # This case handles the stringified `"Optional[A]"` syntax
        # This case handles the stringified `"A | None"` syntax
        case ast.Constant():
            return "Optional[" in annotation.value or "| None" in annotation.value

        case _:
            return False


class UnsetInputTypesPlugin(Plugin):
    """Ariadne plugin to handle Strawberry UNSET types in input types.

    Version 0.7.1 of Ariadne codegen handles UNSET types nicely for arguments in
    queries, but not within input fields, thus fields that should have been
    optional and defaulting to UNSET are instead optional, but required due to
    having no default value.

    It is not possible to simply set the value to UNSET when constructing the
    objects as Pydantic refuses to parse the UNSET variable as an optional
    Employee or similar.

    Future versions of Ariadne seemingly handles UNSET in input types correctly,
    by utilizing the `exclude_unset` feature in Pydantic 2, however we are bound
    to Pydantic 1, as MO cannot be upgraded due to FastAPI being a blocker.

    Thus for now we are at an impasse that this plugin attempts to solve.

    The plugin modifies the generated GraphQL client, such that input types which
    are UNSET in the GraphQL schema can be either their original type or the
    UnsetType while defaulting to the UNSET type.

    This ensures that the fields are no longer required, and that the UNSET
    behavior works as expected, as Ariadne 0.7.1 correctly strips UNSET values
    from the payloads within the `execute` method of the client, as used by
    arguments in queries.

    See:
    * https://github.com/mirumee/ariadne-codegen/issues/121
    * https://github.com/mirumee/ariadne-codegen/pull/128
    * https://github.com/mirumee/ariadne-codegen/issues/165
    For details.

    The plugin can be used by inserting it into the `tool.ariadne-codgen` section
    of the `pyproject.toml` file, as shown below:
    ```toml
        [tool.ariadne-codegen]
        target_package_name = "autogenerated_graphql_client"
        ...
        plugins = [
            # Return values directly when only a single top field is requested
            "ariadne_codegen.contrib.shorter_results.ShorterResultsPlugin",
            # Add UNSET and UnsetType to generated input types
            "fastramqpi.ariadne.UnsetInputTypesPlugin",  # <-- This
        ]
    ```

    This plugin can be removed once we utilize a newer version of Ariadne codegen.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # This plugin probably does not work on pydantic v2 / newer ariadne versions.
        if version("ariadne-codegen") != "0.7.1":  # pragma: no cover
            raise AssertionError(
                "The UnsetInputTypesPlugin only works on Ariadne 0.7.1"
            )

    def generate_inputs_module(self, module: ast.Module) -> ast.Module:
        # This imports the UnsetType and UNSET value atop the input_types module
        unset_imports = [
            ast.ImportFrom(
                level=1, module="base_model", names=[ast.alias("UnsetType")]
            ),
            ast.ImportFrom(level=1, module="base_model", names=[ast.alias("UNSET")]),
        ]

        module.body = unset_imports + module.body
        return module

    def generate_input_field(
        self,
        field_implementation: ast.AnnAssign,
        input_field: GraphQLInputField,
        field_name: str,
    ) -> ast.AnnAssign:
        # Only fields without default values in the GraphQL schema default to UNSET
        if input_field.default_value != Undefined:
            return field_implementation

        # Only optional fields can have difference between UNSET and null
        optional = _is_ast_annotation_optional(field_implementation.annotation)
        if not optional:
            return field_implementation

        # This occurs for fields generated with aliasses, such as `from` being
        # generated as `_from: datetime = Field(alias="from")`, thus being given
        # a default value even if one does not exist in the GraphQL schema.
        # We intentionally ignore these cases, as it only really affects `from`
        # and as we are in the process of replacing `from`/`to` with `start`/`end`.
        if field_implementation.value is not None:
            return field_implementation

        new_annotation: ast.expr
        # Types in the schema may either be real types or stringified ones
        # i.e. either `Optional[MyType]` or `"Optional[MyType]"`.
        # The below code has a branch for each case.
        if (
            isinstance(field_implementation.annotation, ast.Name)
            and '"' in field_implementation.annotation.id
        ) or (isinstance(field_implementation.annotation, ast.Constant)):
            # This handles the stringified case, by stripping the quotes,
            # adding our new type and readding the quotes, such that
            # `"Optional[MyType]"` becomes `"Optional[MyType] | UnsetType"`.
            annotation = field_implementation.annotation
            annotation_value: str = cast(
                str,
                (
                    annotation.id
                    if isinstance(annotation, ast.Name)
                    else annotation.value
                ),
            )
            unquoted_id = annotation_value.strip('"')
            new_annotation = ast.Name(f'"{unquoted_id} | UnsetType"')
        else:
            # This handles real types, by introducing a BinOp.
            # The BinOp ast node is; `left OP right`, where:
            # * `left` is the old annotation,
            # * `op` is `BitOr` aka `|` and
            # * `b` is our `UnsetType`
            # Such that `Optional[MyType]` becomes `Optional[MyType] | UnsetType`.
            new_annotation = ast.BinOp(
                left=field_implementation.annotation,
                op=ast.BitOr(),
                right=ast.Name("UnsetType"),
            )

        # This creates an assignment ast node, i.e. `target: annotation = value`,
        # where we update the annotation to the newly calculated one, and the value
        # is set to `UNSET`.
        return ast.AnnAssign(
            target=field_implementation.target,
            annotation=new_annotation,
            value=ast.Name("UNSET"),
            simple=field_implementation.simple,
        )

    def generate_inputs_code(self, generated_code: str) -> str:
        plugin_header = "# This file has been modified by the UnsetInputTypesPlugin\n"
        return plugin_header + generated_code



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from asyncio import gather
from asyncio import Semaphore
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Awaitable
from typing import Callable
from typing import cast
from typing import List
from typing import TypeVar

ReturnType = TypeVar("ReturnType")
WithConcurrencyFunction = Callable[..., Awaitable[ReturnType]]


def with_concurrency(
    parallel: int,
) -> Callable[[WithConcurrencyFunction], WithConcurrencyFunction]:
    """Decorator which limits concurrency.

    Example:
        ```Python
        @with_concurrency(5)
        async def intensive_task(i: int) -> None:
            ...
            return i

        tasks = list(map(intensive_task, range(1000)))

        # Runs 5 intensive_tasks in parallel
        await asyncio.gather(*tasks)
        ```

    Args:
        parallel: The number of concurrent tasks being executed (must be positive).

    Raises:
        TypeError: if parallel has the wrong type.
        ValueError: if parallel is negative.

    Returns:
        Decorator function to limit concurrency.
    """
    semaphore = Semaphore(parallel)

    @asynccontextmanager
    async def limited_concurrency() -> AsyncGenerator[None, None]:
        async with semaphore:
            yield

    return limited_concurrency()


async def gather_with_concurrency(
    parallel: int, *tasks: Awaitable[ReturnType]
) -> List[ReturnType]:
    """Asyncio gather, but with limited concurrency.

    Example:
        ```Python
        async def intensive_task(i: int) -> None:
            ...
            return i

        tasks = list(map(intensive_task, range(1000)))

        # Runs 1000 intensive_tasks in parallel
        await asyncio.gather(*tasks)
        # Runs at most 5 intensive_tasks in parallel, through all 1000
        await gather_with_concurrency(5, *tasks)
        ```

    Args:
        parallel: The number of concurrent tasks being executed (must be positive).
        tasks: List of tasks to execute.

    Raises:
        TypeError: if parallel has the wrong type.
        ValueError: if parallel is negative.

    Returns:
        List of return values from awaiting the tasks.
    """

    @with_concurrency(parallel)
    async def resolver_task(task: Awaitable[ReturnType]) -> ReturnType:
        return await task

    return cast(List[ReturnType], await gather(*map(resolver_task, tasks)))



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from prometheus_client import Gauge


# If the integration exposes a trigger endpoint, it should make sure to call
# `dipex_last_success_timestamp.set_to_current_time()` to instrument when it last
# successfully ran.
dipex_last_success_timestamp = Gauge(
    name="dipex_last_success_timestamp",
    documentation="When the integration last successfully ran.",
    unit="seconds",
)



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-few-public-methods
"""This module contains the MO specific AMQPSystem."""
from collections.abc import Callable
from datetime import datetime
from typing import Annotated
from typing import Any
from typing import Literal
from uuid import UUID

from fastapi import Depends
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel

from .abstract import AbstractAMQPSystem
from .abstract import AbstractPublishMixin
from .abstract import AbstractRouter
from .depends import get_payload_as_type
from .depends import get_routing_key
from .utils import CallbackType

# >>> print("\n".join(((f"\"{'.'.join(x)}\"," for x in itertools.product(["employee", "org_unit", "*"], ["address", "association", "employee", "engagement", "it", "kle", "leave", "manager", "owner", "org_unit", "related_unit", "role", "*"], ["create", "edit", "terminate", "refresh", "*"])))))  # noqa: E501
_MORoutingKeyOld = Literal[
    "employee.address.create",
    "employee.address.edit",
    "employee.address.terminate",
    "employee.address.refresh",
    "employee.address.*",
    "employee.association.create",
    "employee.association.edit",
    "employee.association.terminate",
    "employee.association.refresh",
    "employee.association.*",
    "employee.employee.create",
    "employee.employee.edit",
    "employee.employee.terminate",
    "employee.employee.refresh",
    "employee.employee.*",
    "employee.engagement.create",
    "employee.engagement.edit",
    "employee.engagement.terminate",
    "employee.engagement.refresh",
    "employee.engagement.*",
    "employee.it.create",
    "employee.it.edit",
    "employee.it.terminate",
    "employee.it.refresh",
    "employee.it.*",
    "employee.kle.create",
    "employee.kle.edit",
    "employee.kle.terminate",
    "employee.kle.refresh",
    "employee.kle.*",
    "employee.leave.create",
    "employee.leave.edit",
    "employee.leave.terminate",
    "employee.leave.refresh",
    "employee.leave.*",
    "employee.manager.create",
    "employee.manager.edit",
    "employee.manager.terminate",
    "employee.manager.refresh",
    "employee.manager.*",
    "employee.owner.create",
    "employee.owner.edit",
    "employee.owner.terminate",
    "employee.owner.refresh",
    "employee.owner.*",
    "employee.org_unit.create",
    "employee.org_unit.edit",
    "employee.org_unit.terminate",
    "employee.org_unit.refresh",
    "employee.org_unit.*",
    "employee.related_unit.create",
    "employee.related_unit.edit",
    "employee.related_unit.terminate",
    "employee.related_unit.refresh",
    "employee.related_unit.*",
    "employee.role.create",
    "employee.role.edit",
    "employee.role.terminate",
    "employee.role.refresh",
    "employee.role.*",
    "employee.*.create",
    "employee.*.edit",
    "employee.*.terminate",
    "employee.*.refresh",
    "employee.*.*",
    "org_unit.address.create",
    "org_unit.address.edit",
    "org_unit.address.terminate",
    "org_unit.address.refresh",
    "org_unit.address.*",
    "org_unit.association.create",
    "org_unit.association.edit",
    "org_unit.association.terminate",
    "org_unit.association.refresh",
    "org_unit.association.*",
    "org_unit.employee.create",
    "org_unit.employee.edit",
    "org_unit.employee.terminate",
    "org_unit.employee.refresh",
    "org_unit.employee.*",
    "org_unit.engagement.create",
    "org_unit.engagement.edit",
    "org_unit.engagement.terminate",
    "org_unit.engagement.refresh",
    "org_unit.engagement.*",
    "org_unit.it.create",
    "org_unit.it.edit",
    "org_unit.it.terminate",
    "org_unit.it.refresh",
    "org_unit.it.*",
    "org_unit.kle.create",
    "org_unit.kle.edit",
    "org_unit.kle.terminate",
    "org_unit.kle.refresh",
    "org_unit.kle.*",
    "org_unit.leave.create",
    "org_unit.leave.edit",
    "org_unit.leave.terminate",
    "org_unit.leave.refresh",
    "org_unit.leave.*",
    "org_unit.manager.create",
    "org_unit.manager.edit",
    "org_unit.manager.terminate",
    "org_unit.manager.refresh",
    "org_unit.manager.*",
    "org_unit.owner.create",
    "org_unit.owner.edit",
    "org_unit.owner.terminate",
    "org_unit.owner.refresh",
    "org_unit.owner.*",
    "org_unit.org_unit.create",
    "org_unit.org_unit.edit",
    "org_unit.org_unit.terminate",
    "org_unit.org_unit.refresh",
    "org_unit.org_unit.*",
    "org_unit.related_unit.create",
    "org_unit.related_unit.edit",
    "org_unit.related_unit.terminate",
    "org_unit.related_unit.refresh",
    "org_unit.related_unit.*",
    "org_unit.role.create",
    "org_unit.role.edit",
    "org_unit.role.terminate",
    "org_unit.role.refresh",
    "org_unit.role.*",
    "org_unit.*.create",
    "org_unit.*.edit",
    "org_unit.*.terminate",
    "org_unit.*.refresh",
    "org_unit.*.*",
    "*.address.create",
    "*.address.edit",
    "*.address.terminate",
    "*.address.refresh",
    "*.address.*",
    "*.association.create",
    "*.association.edit",
    "*.association.terminate",
    "*.association.refresh",
    "*.association.*",
    "*.employee.create",
    "*.employee.edit",
    "*.employee.terminate",
    "*.employee.refresh",
    "*.employee.*",
    "*.engagement.create",
    "*.engagement.edit",
    "*.engagement.terminate",
    "*.engagement.refresh",
    "*.engagement.*",
    "*.it.create",
    "*.it.edit",
    "*.it.terminate",
    "*.it.refresh",
    "*.it.*",
    "*.kle.create",
    "*.kle.edit",
    "*.kle.terminate",
    "*.kle.refresh",
    "*.kle.*",
    "*.leave.create",
    "*.leave.edit",
    "*.leave.terminate",
    "*.leave.refresh",
    "*.leave.*",
    "*.manager.create",
    "*.manager.edit",
    "*.manager.terminate",
    "*.manager.refresh",
    "*.manager.*",
    "*.owner.create",
    "*.owner.edit",
    "*.owner.terminate",
    "*.owner.refresh",
    "*.owner.*",
    "*.org_unit.create",
    "*.org_unit.edit",
    "*.org_unit.terminate",
    "*.org_unit.refresh",
    "*.org_unit.*",
    "*.related_unit.create",
    "*.related_unit.edit",
    "*.related_unit.terminate",
    "*.related_unit.refresh",
    "*.related_unit.*",
    "*.role.create",
    "*.role.edit",
    "*.role.terminate",
    "*.role.refresh",
    "*.role.*",
    "*.*.create",
    "*.*.edit",
    "*.*.terminate",
    "*.*.refresh",
    "*.*.*",
]

_MORoutingKey = Literal[
    "address",
    "association",
    "class",
    "engagement",
    "facet",
    "itsystem",
    "ituser",
    "kle",
    "leave",
    "manager",
    "manager",
    "org_unit",
    "owner",
    "person",
    "related_unit",
    "role",
]

MORoutingKey = Annotated[_MORoutingKey | _MORoutingKeyOld, Depends(get_routing_key)]


class _PayloadType(BaseModel):
    """MO AMQP message format.

    Args:
        uuid: The UUID of the org-unit or employee.
        object_uuid: The UUID of the object itself.
        time: Time when MO emitted the message.
    """

    uuid: UUID
    object_uuid: UUID
    time: datetime


PayloadType = Annotated[_PayloadType, Depends(get_payload_as_type(_PayloadType))]
# The new MO amqp system payloads are just uuids
PayloadUUID = Annotated[UUID, Depends(get_payload_as_type(UUID))]


class MORouter(AbstractRouter):
    """MO specific Router.

    Has specifically tailored `register` methods, which utilize the MO AMQP
    routing-key structure and payload format.
    """

    def register(
        self, routing_key: MORoutingKey, dependencies: list[Any] | None = None
    ) -> Callable[[CallbackType], CallbackType]:  # pragma: no cover
        """Get a decorator for registering callbacks.

        See the documentation in AbstractRouter for more information.

        Args:
            routing_key: The routing key to bind messages for.

        Returns:
            A decorator for registering a function to receive callbacks.
        """
        return self._register(routing_key, dependencies)


class MOPublishMixin(AbstractPublishMixin):
    """MO specific PublishMixin.

    Has a specifically tailored `publish_message` method, which utilize the MO AMQP
    routing-key structure and payload format.
    """

    async def publish_message(
        self, routing_key: MORoutingKey, payload: PayloadType
    ) -> None:
        """Publish a message to the given service.object.request tuple.

        Args:
            routing_key: The routing key to send the message to.
            payload: The message payload.
        """
        payload_obj = jsonable_encoder(payload)
        await self._publish_message(str(routing_key), payload_obj)


class MOAMQPSystem(AbstractAMQPSystem[MORouter], MOPublishMixin):
    """MO specific AMQPSystem."""

    router_cls = MORouter



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# flake8: noqa
import re
from functools import lru_cache
from typing import Pattern

from hypothesis import strategies as st

# Strategies


def not_from_regex(str_pat: str) -> st.SearchStrategy:
    """Generates strings that do not match the `str_pat` regex pattern.

    Inverted <a href="https://hypothesis.readthedocs.io/en/latest/data.html#hypothesis.strategies.from_regex">`hypothesis.strategies.from_regex`</a> search strategy.

    *Note: The `str_pat` regex pattern is compiled and cached internally.*

    Args:
        str_pat: The regex pattern to generate strings that does not match.

    Returns:
        Strategy that generates strings that does not match the `str_pat` regex pattern.
    """

    @lru_cache
    def cached_regex(str_pat: str) -> Pattern:
        return re.compile(str_pat)

    regex = cached_regex(str_pat)
    return st.text().filter(lambda s: regex.match(s) is None)
